/*=================================================================
 *
 * mexMSig.c (C/C++ with MEX)
 *
 * A Mercer Sigmoid kernel 
 * Copyright 2014, Andre M. Carrington.
 *
 * The calling syntax is:
 * [K] = mexMSig(U,V,b,d)
 *
 * This kernel matrix compares every row vector in U (a p-dimensional data 
 * point) with every row vector in V (another p-dimensional data point).
 * 
 * U is Mxp, M instances of horizontal p-vectors
 * V is Nxp, N instances of horizontal p-vectors
 * K is MxN, MxN reals
 *
 *=================================================================*/
#include <math.h>
//#include <stdio.h>
#include "mex.h"
#include "matrix.h"

/* Input Arguments */
#define	U_IN	prhs[0]
#define	V_IN	prhs[1]
#define	b_IN	prhs[2]
#define	d_IN	prhs[3]

/* Output Arguments */
#define	K_OUT	plhs[0] /*YP_OUT*/

static void mexMSig(
		   double	*K, /* kernel matrix output*/
		   double	*U, /* kernel matrix inputs*/
 		   double	*V,
 		   double	b,  /* kernel hyperparms: b = horizontal scalar */
 		   double	d,  /*                    d = horizontal shifter */
           int      M,  /* kernel matrix dimensions */
           int      N,
           int      p   /* vector (data point) dimension/length */
		   )
{
    
    double ku[M][p];
    double kv[p][N];

    // phi_no_norm = @(x) (tanh((x-d)/b)+1);
    // for m=1:M
    //   ku(m,:) = phi_no_norm(U(m,:));  % horizontal vector
    // end
    for(int i=0; i<M; i++) {           /* for each row i in K */
        for (int q=0,iq; q<p; q++) {      /* for each p-dimensional element */
          iq = q*M+i;                    
          ku[i][q] = tanh((U[iq]-d)/b);
        }
    }        
    
    // for n=1:N
    //   kv(:,n) = phi_no_norm(V(n,:)'); % vertical vector
    // end
    for(int j=0; j<N; j++) {      /* for each col j in K */
        for (int q=0,qj; q<p; q++) {      /* for each p-dimensional element */
          qj = q*N+j;                    
          kv[q][j] = tanh((V[qj]-d)/b);
        }
    }
    
    /* K is M x N */
    for(int i=0; i<M; i++) {           /* for each row i in K */
      for(int j=0,ij; j<N; j++) {      /* for each col j in K */
        ij = j*M+i;                    
        for (int q=0; q<p; q++) {      /* for each p-dimensional element */
          K[ij] = K[ij] + ku[i][q]*kv[q][j];
        }
        K[ij] = K[ij] / p;
      }
    }    
    return;
}

void mexFunction( int nlhs, mxArray *plhs[], 
		  int nrhs, const mxArray*prhs[] )
     
{ 
    double *K, *U, *V, *b, *d;
    size_t um,p1,vm,p2; 
    
    /* Check for proper number of arguments */
    if (nrhs != 4) { 
	    mexErrMsgIdAndTxt( "MATLAB:mexMSig:invalidNumInputs",
                "Four input arguments required."); 
    } else if (nlhs > 1) {
	    mexErrMsgIdAndTxt( "MATLAB:mexMSig:maxLHS",
                "Too many output arguments."); 
    } 
    
    /* Check the dimensions of U and V. */ 
    um = mxGetM(U_IN); 
    p1 = mxGetN(U_IN);
    vm = mxGetM(V_IN); 
    p2 = mxGetN(V_IN);
    
    if (p1 != p2) { 
	    mexErrMsgIdAndTxt( "MATLAB:mexMSig:invalidDimensions",
                "Inputs must have the same number of columns."); 
    } 
    
    K_OUT = mxCreateDoubleMatrix(um, vm, mxREAL); /* all elements 0 */
    *mxGetPr(K_OUT) = 0;

    /* Assign pointers to the various parameters */ 
    K = mxGetPr(K_OUT);
    U = mxGetPr(U_IN); 
    V = mxGetPr(V_IN);
    b = mxGetPr(b_IN);
    d = mxGetPr(d_IN);
        
    /* Do the actual computations in a subroutine */
    mexMSig(K,U,V,*b,*d,(int)um,(int)vm,(int)p1);
    return;
    
}
